/*
 * This file is part of the SPiC RobotArm (https://gitlab.cs.fau.de/i4/spic/robotarm).
 * Copyright (c) 2019 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdbool.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/sleep.h>
#define BAUD 38400
#include <util/setbaud.h>

#define SERVOS 4

#define UNSET 65535

static const uint16_t timer_max = 39999;

static const struct {
	volatile uint8_t * const ddr;
	volatile uint8_t * const port;
	const uint8_t pin;
	const uint8_t duration; // operation time factor (x 20 ms)
	const uint16_t min;     // duty cycle minimum
	const uint16_t max;     // duty cycle maximum
} servo[SERVOS] = {
	// Configuration with adjusted duty cycles
	{&DDRD, &PORTD, PD4, 100, 1500, 5200}, // middle
	{&DDRD, &PORTD, PD5, 100, 1200, 4700}, // left
	{&DDRD, &PORTD, PD6, 100, 1300, 4800}, // right
	{&DDRD, &PORTD, PD7, 100, 1350, 4850}, // claw
};

static volatile uint16_t duty_cycle[SERVOS]= { 0 };

static volatile bool relax = true;

static volatile char respond = '@';

// Timer interrupt on compare match
ISR(TIMER1_COMPA_vect) {
	static uint16_t duty_cycle_cache[SERVOS] = { 0 }; // Cache for full cycle
	static uint16_t timeout[SERVOS] = { 0 }; // timeout of servo
	uint16_t current_counter = OCR1A;
	uint16_t next_counter = timer_max;

	if (current_counter >= timer_max){
		TCNT1 = 0;
		if (relax){
			for (uint8_t i = 0; i < SERVOS; i++){
				if (duty_cycle[i] != UNSET){
					duty_cycle_cache[i] = duty_cycle[i];
					duty_cycle[i] = UNSET;
					timeout[i] = servo[i].duration;
				}
				if (timeout[i] > 0){
					timeout[i]--;
					if (duty_cycle_cache[i] > 0){
						*(servo[i].port) |= (1 << servo[i].pin);
						if (duty_cycle_cache[i] < next_counter)
							next_counter = duty_cycle_cache[i];
					}
				}
			}
		} else
			for (uint8_t i = 0; i < SERVOS; i++)
				if (duty_cycle_cache[i] != duty_cycle[i])
					duty_cycle_cache[i] = duty_cycle[i];
	} else {
		for (uint8_t i = 0; i < SERVOS; i++){
			// disable pin
			if (duty_cycle_cache[i] <= current_counter)
				*(servo[i].port) &= ~(1 << servo[i].pin);
			// calculate next stop
			else if (duty_cycle_cache[i] < next_counter)
				next_counter = duty_cycle_cache[i];
		}
	}
	OCR1A = next_counter;
}

// USART Receive interrupt
ISR(USART0_RX_vect){
	static uint8_t id = 0; // Servo number (0-4)
	static uint16_t value = 0; // Angle in degrees (0 - 180)
	static bool raw = false;

	unsigned char nextChar = UDR0;

	// Reset connection
	if (nextChar == '*'){
		id = 0;
		value = 0;
		raw = false;
		relax = true;
		respond = '*';
	}

	// Angle for specified servo (A, B, ...)
	else if (nextChar >= 'A' && nextChar < 'A' + SERVOS){
		id = nextChar - 'A';
		value = 0;
		raw = false;
	}

#ifdef DEBUG
	// Raw Value for specified servo (a, b, ...)
	else if (nextChar >= 'a' && nextChar < 'a' + SERVOS){
		id = nextChar - 'a';
		value = 0;
		raw = true;
	}
#endif

	// Value
	else if (nextChar >= '0' && nextChar <= '9')
		value = value * 10 + (nextChar - '0');

	// Relax mode
	else if (nextChar == '~'){
		relax = true;
		respond = nextChar;
	}

	// Permanent force on servos
	else if (nextChar == '-'){
		relax = false;
		respond = nextChar;
	}

	// Ping
	else if (nextChar == '.')
		respond = nextChar; // Pong

	// Action
	else if (nextChar == '\n'){
		if (raw && value < timer_max){
			duty_cycle[id] = value;
			respond = '#'; // Raw value ok
		}
		else if (!raw && value <= 180){
			duty_cycle[id] = ((uint32_t)(servo[id].max - servo[id].min) * value) / 180 + servo[id].min;
			respond = '+'; // Angle ok
		}
		else
			respond = '!'; // Error
		value = 0;
	}

	// Invalid symbol
	else
		respond = '?';
}

void main(void){
	// Initialize Timer 1
	TCCR1A = 0;
	// Prescaler 8
	TCCR1B = (1 << CS11);
	// activate timer 1 compate interrupt
	TIMSK1 |= (1<<OCIE1A);
	// set counter to zero
	TCNT1 = 0;
	// set compare value to maximum
	OCR1A = timer_max;

	// USART
	// Set baud rate

	// USART
	UBRR0H = UBRRH_VALUE;
	UBRR0L = UBRRL_VALUE;
#if USE_2X
	UCSR0A |= (1 << U2X0);
#else
	UCSR0A &= ~(1 << U2X0);
#endif
	// set 8 bit frame
	UCSR0C = (1<<UCSZ01)|(1<<UCSZ00);
	// enable transmission, reception and receive interrupt
	UCSR0B |= (1<<RXEN0)|(1<<TXEN0)|(1<<RXCIE0);

	// initialize pins
	 for (uint8_t i = 0; i < SERVOS; i++) {
		*(servo[i].ddr) |= (1 << servo[i].pin);
		*(servo[i].port) &= ~(1 << servo[i].pin);
	}

	// enable interrupts
	sei();

	sleep_enable();

	while(1) {
		sleep_cpu();
		if (respond != '\0'){
			while(!(UCSR0A&(1<<UDRE0)));
			UDR0 = respond;
			respond = '\0';
		}
	}
}
