/*
 * This file is part of the SPiC RobotArm (https://gitlab.cs.fau.de/i4/spic/robotarm).
 * Copyright (c) 2019 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <unistd.h>

#include "armconfig.h"
#include "arm.h"
#include "com.h"
#include "coord.h"

// Storage for current arm position
static double xCur = XINIT;
static double yCur = YINIT;
static double zCur = ZINIT;

extern bool verbose;

/*! \brief Inverse kinematics: determine the joint parameters by the cartesian coordinates
 * \param x requested X position
 * \param y requested Y position
 * \param z requested Z position
 * \param base pointer for resulting base joint angle (in degrees)
 * \param shoulder pointer for resulting shoulder joint angle (in degrees)
 * \param elbow pointer for resulting elbow joint angle (in degrees)
 * \return `true` if robot arm is able to move to the requested position
 */
static bool solve(double x, double y, double z, double *base, double *shoulder, double *elbow){
	// top-down view
	polar_t top = cart2polar(y, x);

	// take width of gripper and base into account
	if (top.p < DW){
		return false;
	} else {
		top.p -= DW;
	}

	// side view (take height into account)
	polar_t arm = cart2polar(top.p, z - DH);

	// solve arm inner angles as required
	double beta = cosangle(DU, arm.p, DF);
	double gamma = cosangle(DU, DF, arm.p);

	// invalid result --> unsolvable
	if(isnan(beta) || isnan(gamma)){
		return false;
	}

	// Servo angles
	double beta_total = beta + arm.phi;
	double beta_servo = M_PI - beta_total;
	double gamma_servo = gamma - (M_PI/2 - beta_total);

	if (verbose)
		fprintf(stderr,"\x1b[2mα=%.2lf°, β=%.2lf° (βₛ=%.2lf°), γ=%.2lf° (γₛ=%.2lf°), φ=%.2lf°\x1b[0m\n", rad2deg(top.phi), rad2deg(beta), rad2deg(beta_servo), rad2deg(gamma), rad2deg(gamma_servo), rad2deg(arm.phi));

	// solve for servo angles from horizontal
	*base = rad2deg(top.phi);
	*shoulder = rad2deg(beta_servo);
	*elbow = rad2deg(gamma_servo);

	return true;
}

bool armJointMove(enum armJoint id, double value){
	if (!isfinite(value)){
		return false;
	}

	int i = (int)round(value);

	if (i < 0)
		i = 0;

	if (i > 180)
		i = 180;

	comSend(id);

	if (i >= 100){
		comSend('1');
		i %= 100;
	}
	comSend('0' + (i / 10));
	comSend('0' + (i % 10));
	comSend('\n');

	return comReceive() == '+';
}

bool armMove(double x, double y, double z, enum armMovement movement){
	double base, shoulder, elbow;
	if (solve(x, y, z, &base, &shoulder, &elbow)) {
		switch(movement) {
			case DEFAULT:
			case INSTANT:
				xCur = x; yCur = y; zCur = z;

				if (verbose)
					fprintf(stderr,"\x1b[2mMove to [%.2lf,%.2lf,%.2lf]\x1b[0m\n", x, y ,z);

				return armJointMove(BASE, base) &&
				       armJointMove(SHOULDER, shoulder) &&
				       armJointMove(ELBOW, elbow);

			case SMOOTH:
				{
					double x0 = xCur;
					double y0 = yCur;
					double z0 = zCur;
					double dist = sqrt((x0-x)*(x0-x)+(y0-y)*(y0-y)+(z0-z)*(z0-z));
					for (int i = 0; i < dist; i+= STEPSIZE) {
						if (!armMove(x0 + (x-x0)*i/dist, y0 + (y-y0) * i/dist, z0 + (z-z0) * i/dist, INSTANT)){
							return false;
						}
						usleep(DELAY);
					}
					return armMove(x, y, z, INSTANT);
				}

			case TEST:
				fprintf(stderr, "[%.2lf,%.2lf,%.2lf] is valid!\n", x, y, z);
				return true;
		}
	}
	fprintf(stderr, "Not able to move to [%.2lf,%.2lf,%.2lf]!\n", x, y, z);
	return false;
}

bool armGrip(double percentage){
	if (isfinite(percentage)){
		if (percentage > 100.0){
			percentage = 100.0;
		} else if (percentage < 0.0){
			percentage = 0.0;
		}
		return armJointMove(GRIPPER, percentage * GRIPPERMAX / 100.0);
	}
	return false;
}

bool armPing(){
	// send ping
	comSend('.');
	// check answer
	return comReceive() == '.';
}

bool armRelax(bool relax){
	char key = relax ? '~' : '-';
	// send ping
	comSend(key);
	// check answer
	return comReceive() == key;
}

bool armReset(){
	// send reset
	comSend('*');
	// wait for correct answer
	while(comReceive() != '*');
	fprintf(stderr, "Connection reset!\n");

	// move to initial position
	return armMove(XINIT, YINIT, ZINIT, INSTANT) && armJointMove(GRIPPER, 0);
}

bool armCalibrate(){
	return armJointMove(BASE, 90) && usleep(300) == 0 &&
	       armJointMove(SHOULDER, 90) && usleep(300) == 0 &&
	       armJointMove(ELBOW, 90) && usleep(300) == 0 &&
	       armJointMove(GRIPPER, 25);
}
