/*
 * This file is part of the SPiC RobotArm (https://gitlab.cs.fau.de/i4/spic/robotarm).
 * Copyright (c) 2019 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ARM_H
#define ARM_H

#include <stdbool.h>

/*! \brief Robot arm joint identification */
enum armJoint {
	BASE = 'A',
	SHOULDER = 'B',
	ELBOW = 'C',
	GRIPPER = 'D'
};

/*! \brief Send a joint movement
 * \param id joint (servo) identification
 * \param value requested target degree
 * \return `true` if movement instruction was accepted by the robot arm
 */
bool armJointMove(enum armJoint id, double value);

/*! \brief types of arm movement */
enum armMovement {
	DEFAULT = 0,  ///< default type is INSTANT
	INSTANT = 1,  ///< move to position (ANY path)
	SMOOTH = 2,   ///< move to position (smooth in a linear path)
	TEST = 3,     ///< just check if movement is valid (don't move)
};

/*! \brief Move the robot arm
 *  \param x X position in cartesian coordinates (millimeters)
 *  \param y Y position in cartesian coordinates (millimeters)
 *  \param z Z position in cartesian coordinates (millimeters)
 *  \param movement type of arm movement
 *  \return `true` if valid coordinates
 */
bool armMove(double x, double y, double z, enum armMovement movment);


/*! \brief Open/close the gripper
 *  \param percentage grippers grip (100% = closed)
 *  \return `true` if command successfully sent
 */
bool armGrip(double percentage);

/*! \brief Ping robot arm
 * \return `true` if correct response was received
 */
bool armPing();

/*! \brief Reset robot arm communication and move to initial position
 * \return `true` if reset was successful
 */
bool armReset();

/*! \brief Calibrate servos for building
 * \return `true` if calibration was successfully sent
 */
bool armCalibrate();

/*! \brief Relax robot arm servos after timeout
 * \param relax `true` will let the servos relax after timeout,
 *              `false` will always keep force
 * \return `true` if command was successful
 */
bool armRelax(bool relax);

#endif
