/*
 * This file is part of the SPiC RobotArm (https://gitlab.cs.fau.de/i4/spic/robotarm).
 * Copyright (c) 2019 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "com.h"

static speed_t baud = B38400;
static struct termios spicboard;

static int fd = -1;

extern bool verbose;

void comInit(){
	spicboard.c_cflag = CREAD | CLOCAL | CS8;  // use 8n1 connection by default
	spicboard.c_cc[VMIN] = 1;  // read should block
	spicboard.c_cc[VTIME] = 5;  // 0.5s read timeout
}

bool comTimeout(const char * args){
	char * endptr;
	errno = 0;
	unsigned long val = strtoul(args, &endptr, 10);
	if (errno != 0 || endptr == optarg){
		fprintf(stderr, "Not able to parse timeout '\x1b[1m%s\x1b[0m'!\n", args);
		return false;
	}
	spicboard.c_cc[VTIME] = val;
	return true;
}


bool comBaudrate(const char * args){
	char * endptr;
	errno = 0;
	unsigned long val = strtoul(args, &endptr, 10);
	if (errno != 0 || endptr == optarg){
		fprintf(stderr, "Not able to parse baud rate '\x1b[1m%s\x1b[0m'!\n", args);
		return false;
	} else {
		switch (val){
			case 300: baud = B300; break;
			case 600: baud = B600; break;
			case 1200: baud = B1200; break;
			case 2400: baud = B2400; break;
			case 4800: baud = B4800; break;
			case 9600: baud = B9600; break;
			case 19200: baud = B19200; break;
			case 38400: baud = B38400; break;
			case 57600: baud = B57600; break;
			case 115200: baud = B115200; break;
			default:
				fprintf(stderr, "Unsupported baud rate \x1b[4m%lu\x1b[0m!\n", val);
				return false;
		}
	}
	return true;
}

bool comSettings(const char * args){
	// Reset to basic
	spicboard.c_cflag = CREAD | CLOCAL;

	// Datenbits
	switch (args[0]){
		case '5': spicboard.c_cflag |= CS5; break;
		case '6': spicboard.c_cflag |= CS6; break;
		case '7': spicboard.c_cflag |= CS7; break;
		case '8': spicboard.c_cflag |= CS8; break;
		default:
			fprintf(stderr, "Unsupported or invalid data bit parameter '\x1b[4m%c\x1b[0m'\n", args[0]);
			return false;
	}

	// Parität
	switch (args[1]){
		case 'E':
		case 'e':
			spicboard.c_cflag |= PARENB ;
			spicboard.c_iflag |= INPCK | ISTRIP;
			break;
		case '0':
		case 'O':
		case 'o':
			spicboard.c_cflag |= PARENB | PARODD;
			spicboard.c_iflag |= INPCK | ISTRIP;
			break;
		case 'N':
		case 'n':
			break;
		default:
			fprintf(stderr, "\nUnsupported or invalid parity parameter '\x1b[4m%c\x1b[0m'\n", args[1]);
			return false;
	}

	// Stoppbits
	switch (args[2]){
		case '1': break;
		case '2': spicboard.c_cflag |= CSTOPB; break;
		default:
			fprintf(stderr, "\rUnsupported or invalid stop bit parameter '\x1b[4m%c\x1b[0m'\n", args[2]);
			return false;
	}

	return true;
}

bool comConnect(const char * device){
	if ((fd = open (device, O_RDWR | O_NOCTTY | O_SYNC)) < 0){
		perror("Not able to open connection to robot arm");
		return false;
	}

	struct termios tty;
	memset (&tty, 0, sizeof tty);
	if (tcgetattr(fd, &tty) != 0){
		perror("Not able to get attributes of terminal");
		return false;
	}

	cfsetospeed (&tty, baud);
	cfsetispeed (&tty, baud);

	if (tcsetattr (fd, TCSANOW, &tty) != 0){
		perror("Not able to set attributes of terminal");
		return false;
	}
	return true;
}


char comReceive(){
	char data;
	errno = 0;
	int n = read(fd, &data, 1);
	if (n == 0){
		fputs("\r\nFehler bei Eingabe -- irgendwas ist total falsch!\n", stderr);
		exit(EXIT_FAILURE);
	} else if (n == -1){
		fprintf(stderr, "\r\n\x1b[2mFehler bei Eingabe: %s\x1b[0m\r\n", strerror(errno));
		exit(EXIT_FAILURE);
	} else if (verbose) {
		fprintf(stderr, "\x1b[35m%c\x1b[0m", data);
	}
	return data;
}

void comSend(const char data){
	errno = 0;
	int n = write(fd, &data, 1);
	if (n == 0){
		fputs("\r\n\x1b[2mFehler bei Ausgabe -- irgendwas ist total falsch!\x1b[0m\r", stderr);
		exit(EXIT_FAILURE);
	} else if (n == -1){
		fprintf(stderr, "\r\n\x1b[2mFehler bei Ausgabe: %s\x1b[0m\r\n", strerror(errno));
		exit(EXIT_FAILURE);
	} else if (verbose){
		fprintf(stderr, "\x1b[36m%c\x1b[0m", data);
	}
}
