/*
 * This file is part of the SPiC RobotArm (https://gitlab.cs.fau.de/i4/spic/robotarm).
 * Copyright (c) 2019 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>

#include "arm.h"
#include "com.h"
#include "parse.h"

static void help(const char * util){
	fprintf(stderr,
	        "\n\x1b[1;4mSPiC RobotArm Controller\x1b[0m\n"
	        "\n"
	        "\x1b[1mUsage:\x1b[0m\n"
	        "    %s"
	        " [\x1b[4m-h\x1b[0m]"
	        " [\x1b[4m-c\x1b[0m]"
	        " [\x1b[4m-r\x1b[0m]"
	        " [\x1b[4m-d\x1b[0;3;4mDEVICE\x1b[0m]"
	        " [\x1b[4m-b\x1b[0;3;4mBAUDRATE\x1b[0m]"
	        " [\x1b[4m-s\x1b[0;3;4mSETTINGS\x1b[0m]"
	        " [\x1b[4m-t\x1b[0;3;4mTIMEOUT\x1b[0m]"
	        " [\x1b[4m-v\x1b[0m]"
	        " [\x1b[4m-e\x1b[0m]"
	        " [\x1b[4mFILE\x1b[0m]"
	        "\n",
	        util
	);
	fputs("\n\x1b[1;2mOptions:\x1b[0m\n"
	      "    \x1b[4m-h\x1b[0m"
	      "             show this help\n"
	      "    \x1b[4m-c\x1b[0m"
	      "             move all servos to the calibration position\n"
	      "    \x1b[4m-r\x1b[0m"
	      "             direct raw serial communication with the microcontroller\n"
	      "    \x1b[4m-d\x1b[0;3;4mDEVICE\x1b[0m"
	      "       Path to serial communication device connected with the RobotArm.\n"
	      "                   Defaults to \x1b[1m/dev/ttyACM0\x1b[0m.\n"
	      "    \x1b[4m-b\x1b[0;3;4mBAUD\x1b[0m"
	      "         Transmission speed of the serial connection, from \x1b[1m300\x1b[0m to \x1b[1m115200\x1b[0m Baud.\n"
	      "                   Defaults to \x1b[1m38400\x1b[0m Baud.\n"
	      "    \x1b[4m-s\x1b[0;3;4mSETTINGS\x1b[0m"
	      "     configure the serial connection parameters\n"
	      "                        \x1b[3mdata-bits\x1b[0m, \x1b[3mparity\x1b[0m and \x1b[3mstop-bits\x1b[0m\n"
	      "                   with parity being \x1b[1me\x1b[0m (even), \x1b[1mo\x1b[0m (odd) oder \x1b[1mn\x1b[0m (none)\n"
	      "                   Those three parameter must all occure in the given order without any delimiter in between.\n"
	      "                   Defaults to \x1b[1m8n1\x1b[0m.\n"
	      "    \x1b[4m-t\x1b[0;3;4mTIMEOUT\x1b[0m"
	      "      set the serial communication timeout (int tenth of a second)\n"
	      "                   Defaults to \x1b[1m5\x1b[0m (0.5s).\n"
	      "    \x1b[4m-v\x1b[0m"
	      "             verbose output (print \x1b[2mmovement calculations\x1b[0m and communication \x1b[35mfrom\x1b[0m and \x1b[36mto\x1b[0m the RobotArm)\n"
	      "    \x1b[4m-e\x1b[0m"
	      "             stop on error (invalid command etc.)\n"
	      "    \x1b[4mFILE\x1b[0m"
	      "           Sequential list of commands (according to the protocol) controlling the RobotArm.\n"
	      "                   A dash (\x1b[1m-\x1b[0m) reads from standard input (default).\n"
	      "\n"
	      "\x1b[1;2mCommands:\x1b[0m\n"
	      "  Arm movement\n"
	      "    \x1b[34m[\x1b[34;1mx\x1b[0;34m,\x1b[34;1my\x1b[0;34m,\x1b[34;1mz\x1b[0;34m]\x1b[0m       position in cartesian coordiantes\n"
	      "    \x1b[34m(\x1b[34;1mr\x1b[0;34m,\x1b[34;1mphi\x1b[34m,\x1b[34;1mz\x1b[0;34m)\x1b[0m     cylindrical coordinates\n"
	      "  Prefix modifier for arm movement\n"
	      "    \x1b[34m?\x1b[0m             test next coordinates (but don't move)\n"
	      "    \x1b[34m-\x1b[0m             move the arm in a smooth linear path\n"
	      "    \x1b[34m#\x1b[0m             move the arm instantly (as fast as possible)\n"
	      "  Additional control\n"
	      "    \x1b[34m{\x1b[34;1mpercentage\x1b[0;34m}\x1b[0m  gripper (\x1b[1m0\x1b[0m = open, \x1b[1m100\x1b[0m = closed)\n"
	      "    \x1b[34m<\x1b[34;1mus\x1b[0;34m>\x1b[0m          delay in microseconds\n"
	      "    \x1b[34m.\x1b[0m             ping the device\n"
	      "    \x1b[34m!\x1b[0m             reset the device (should always performed on start)\n"
	      "    \x1b[34mr\x1b[0m             relax servos (after timeout)\n"
	      "    \x1b[34mf\x1b[0m             permanent force servos (take care of power consumption)\n"
	      "    \x1b[34mq\x1b[0m             quit (End of file or null byte work as well)\n"
	      "  Delimiter\n"
	      "    (semi)colon, tab, space and new line / carriage return\n"
	      "\n",
	      stderr
	);
}

bool verbose = false;

int main(int argc, char *argv[]) {
	// initialize serial communication
	comInit();

	// Default values
	char * device = "/dev/ttyACM0";
	bool calibrate = false;
	bool stopOnError = false;
	bool raw = false;

	// Read options
	int opt;
	while ((opt = getopt(argc, argv, "?hcd:b:rs:t:ve")) != -1) {
		switch (opt) {
			// help
			case '?':
			case 'h':
				help(argv[0]);
				return EXIT_SUCCESS;

			// calibration
			case 'c':
				calibrate = true;
				break;

			// device
			case 'd':
				if (optarg != NULL)
					device = optarg;
				break;

			// baud rate
			case 'b':
				if (!comBaudrate(optarg))
					return EXIT_FAILURE;
				break;

			// raw communication
			case 'r':
				raw = true;
				break;

			// serial connection settings
			case 's':
				if (!comSettings(optarg))
					return EXIT_FAILURE;
				break;

			// serial connection timeout
			case 't':
				if (!comTimeout(optarg))
					return EXIT_FAILURE;
				break;

			// verbose output
			case 'v':
				verbose = true;
				break;

			// stop on error
			case 'e':
				stopOnError = true;
				break;

			// invalid option
			default:
				fprintf(stderr, "Invalid option '%c'!", opt);
				return EXIT_FAILURE;
		}
	}

	// Connect to device
	if (!comConnect(device)){
		fputs("Not able to connect to robot arm!", stderr);
		return EXIT_FAILURE;
	}

	if (calibrate){
		fputs("Calibrating...", stderr);
		return armCalibrate() ? EXIT_SUCCESS : EXIT_FAILURE;
	} else {
		// use command file
		char * in = "-";
		if (optind == argc - 1){
			in = argv[optind];
		}

		// open command file
		errno = 0;
		FILE * instream;
		if (in[0] == '-' && in[1] == '\0'){
			instream = stdin;
		} else {
			instream = fopen(in, "r");
		}
		if (instream == 0){
			perror("Could not open command file!");
			return false;
		}

		bool state = true;
		if (raw){
			int c;
			// raw communication
			while ((c = fgetc(instream)) != EOF){
				comSend(c);
				if (c == '\n')
					putchar(comReceive());
			}
		} else {
			// Parse command file
			state = parse(instream, stopOnError);
		}
		fclose(instream);

		return state ? EXIT_SUCCESS : EXIT_FAILURE;
	}
}
