/*
 * This file is part of the SPiC RobotArm (https://gitlab.cs.fau.de/i4/spic/robotarm).
 * Copyright (c) 2019 by Bernhard Heinloth
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdio.h>
#include <unistd.h>

#include "arm.h"
#include "parse.h"
#include "coord.h"

/*! \brief Helper function to parse an integer number (by hand)
 *  Spaces and tabs are allowed before/after the value.
 *  \param stream input stream
 *  \param val pointer to store the number (as floating point value)
 *  \param del expected delimiter to stop parsing
 *  \return `true` if successfully parsed the number and stopped by delimiter
 */
static bool parseNum(FILE *stream, double * val, char del){
	int c;
	int64_t v = 0;
	enum {
		NEGATIVE = -1,
		NONE = 0,
		POSITIVE = 1,
	} sign = NONE;
	bool intStart = false;
	bool intEnd = false;

	while ((c = fgetc(stream)) != EOF){
		// Parse Integer value
		if (!intEnd && c >= '0' && c <= '9'){
			intStart = true;
			v *= 10;
			v += c - '0';
		}
		// Negative sign symbol before integer
		else if (!intStart && c == '-' && sign == NONE)
			sign = NEGATIVE;

		// optional positive sign before integer
		else if (!intStart && c == '+' && sign == NONE)
			sign = POSITIVE;

		// stop at delimiter
		else if (c == del){
			*val = v * (sign == NEGATIVE ? -1.0 : 1.0);
			return true;
		}

		// ignore space and tab
		else if (c == ' ' || c == '\t'){
			// No numbers after space/tab
			if (intStart)
				intEnd = true;
		}

		// Unexpected character
		else {
			fprintf(stderr, "Unexpected character '%c'...\n", c);
			return false;
		}
	}

	// end of input
	fprintf(stderr, "Missing delimiter '%c'...\n", del);
	return false;
}


bool parse(FILE * stream, bool stopOnError){
	int c;
	enum armMovement type = INSTANT;
	while ((c = fgetc(stream)) != EOF)
		switch (c){

			case '?':
				if (type != DEFAULT)
					fprintf(stderr, "Overriding movement type!\n");
				type = TEST;
				break;

			case '-':
				if (type != DEFAULT)
					fprintf(stderr, "Overriding movement type!\n");
				type = SMOOTH;
				break;

			case '#':
				if (type != DEFAULT)
					fprintf(stderr, "Overriding movement type!\n");
				type = INSTANT;
				break;

			// (r, phi, z)
			case '(':
				{
					double r, phi, z;
					if (   parseNum(stream, &r, ',')    // distance
					    && parseNum(stream, &phi, ',')  // angle (degrees)
					    && parseNum(stream, &z, ')')    // height
					   ){
						// Convert to cartesian (for linear smooth movement)
						cart_t cart = polar2cart(r, deg2rad(phi));
						armMove(cart.x, cart.y, z, type);
						type = DEFAULT;
					} else {
						fprintf(stderr, "Invalid formated cylindrical coordinates!\n");
						if (stopOnError)
							return false;
					}
				}
				break;

			// [x, y, z]
			case '[':
				{
					double x, y, z;
					if (   parseNum(stream, &x, ',')  // X
						&& parseNum(stream, &y, ',')  // Y
						&& parseNum(stream, &z, ']')  // height
					   ){
						armMove(x, y, z, type);
						type = DEFAULT;
					} else {
						fprintf(stderr, "Invalid formated cartesian coordinates!\n");
						if (stopOnError){
							return false;
						}
					}
				}
				break;

			// {g}
			case '{':
				{
					double percentage;
					if (parseNum(stream, &percentage, '}')){
						armGrip(percentage);
					} else {
						fprintf(stderr, "Invalid formated gripper percentage!\n");
						if (stopOnError){
							return false;
						}
					}
				}
				break;

			// {g}
			case '<':
				{
					double delay;
					if (parseNum(stream, &delay, '>')){
						if (delay > 0.0){
							usleep((useconds_t) delay);
						}
					} else  {
						fprintf(stderr, "Invalid formated delay time!\n");
						if (stopOnError){
							return false;
						}
					}
				}
				break;

			// Ping
			case '.':
				if (armPing()){
					fprintf(stderr, "Ping ok!\n");
				} else {
					fprintf(stderr, "Incorrect Ping answer!\n");
					if (stopOnError){
						return false;
					}
				}
				break;

			// Reset
			case '!':
				if (armReset())
					fprintf(stderr, "Moved to initial position!\n");
				else {
					fprintf(stderr, "Not able to move to initial position - abort!\n");
					return false;
				}
				break;

			// Ignore
			case ',':
			case ';':
			case ' ':
			case '\t':
			case '\r':
			case '\n':
				break;

			// Exit
			case '\0':
			case 'q':
				fprintf(stderr, "Exit by request!\n");
				return true;

			// Error
			default:
				fprintf(stderr, "Unexpected character '%c'!\n", c);
				if (stopOnError){
					return false;
				}
		}

	fprintf(stderr, "End of input -- exit!\n");
	return true;
}
